#! /usr/bin/env python
"""
/***************************************************************************
                         imgSeekLite Image database
                             -------------------
    begin                : 03/08/17 20:26:16 rnc
    Time-stamp:            <2003-09-12 22:39:06 rnc>   
    copyright            : (C) 2002 by Ricardo Niederberger Cabral
    email                : nieder|at|mail.ru

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ***************************************************************************/
"""

import random                       # for generating unique id's
import os,sys
import traceback
import md5

if os.name in ["nt","dos"]:            
    import _imgdb as imgdb                
else:
    import imgdb as imgdb                


def PrintTB():
    traceback.print_exc()

class ImgDB:
    """ this is the class holding all logical db data
    """
    if os.name in ["nt","dos"]:                   # only check for non-win32, as for win32 it will be all hardcoded    
        supported_ext = ["jpg","jpeg","png","bmp","png","gif","tiff"]    
    else:
        supported_ext = ["jpg","jpeg","png","bmp","png","gif","tiff"]
        
    __dbversion = 1

    def __init__(self,dbfile=None):
        self.__metaDBFile = None
        self.images = {}                  # key is id, value is [path, thumbpath]        
        self.openMetaDB(dbfile or "metadata.db")

        #### Thumbnail stuff
        self.thdir = os.path.expanduser(os.path.join("~",".thumbnails","")) # path for storing thumbnails. Calculate it only once here.
        self.thdir2 = os.path.expanduser(os.path.join("~",".thumbnails","normal","")) # path for storing thumbnails. Calculate it only once here.
        # just to make sure the thumbnail dir exists, so I dont need to be checking it over and over on self.getThumb()
        try:
            if not os.path.exists(self.thdir):
                os.mkdir(self.thdir)
            if not os.path.exists(self.thdir2):
                os.mkdir(self.thdir2)
        except:
            bdir = sys.path[0] + os.sep + os.path.join("share","imgSeekLite","")
            self.thdir = os.path.join(bdir,  ".thumbnails","")
            self.thdir2 = os.path.join(bdir,  ".thumbnails","normal","")
            if not os.path.exists(self.thdir):
                os.mkdir(self.thdir)
            if not os.path.exists(self.thdir2):
                os.mkdir(self.thdir2)            
            
    def __del__(self):
        self.saveMetaDB()

    def reset(self):
        self.images = {}                  # key is id, value is [path, thumbpath]        

    def openMetaDB(self, dbfile):
        assert(dbfile)
        self.__metaDBFile = dbfile
        imgdb.resetdb()
        self.reset()
        imgdb.initDbase()
        
        import marshal                      # for loading and saving img databases
        if os.path.exists(self.__metaDBFile):
            try:            
                f = open(self.__metaDBFile,"rb")
                rdbversion = marshal.load(f)
                self.images = marshal.load(f)
                f.close()
            except:
                PrintTB()
        if not imgdb.loaddb(self.__metaDBFile+".img"):
            print "Error loading image database"
        print "Database opened."
        
    def saveMetaDB(self, dbfile=None):
        if dbfile:
            self.__metaDBFile = dbfile

        import marshal                      # for loading and saving img databases                  
        if not imgdb.savedb(self.__metaDBFile+".img"):
            print "Error saving image signature database"
        try:
            f = open(self.__metaDBFile,"wb")
            marshal.dump(ImgDB.__dbversion,f)
            marshal.dump(self.images,f)
            f.close()
        except:
            PrintTB()
        print "Database saved."
        
        return 1

    def getImgCount(self):
        return len(self.images)

    def getImgMetaData(self,id):
        assert(self.images.has_key(id))
        #TODO1: should i copy data returned here before returning, so external caller can't change internal data ?        
        return self.images[id]
    
    def getRandomID(self):
        try:
            return self.images.keys()[random.randrange(len(self.images.keys()))]
        except:
            return None
    
    def generateImageID(self):
        """ need to return something unique"""
        #TODO: is this random hack safe ? should I hash a resource somehow ? should I check if this ID already exists somewhere ? How much will extra checks slow things down ?
        try:
            a = random.randrange(999999999)
        except:
            try:
                a = random.randrange(99999999)
            except:                
                try:
                    a = random.randrange(9999999)
                except:
                    a = random.randrange(999999)                
        # 1 is a reserved id for default items
        while a==1: a = self.generateImageID()
        return a

    def addFile(self,fname):
        assert(fname)        
        newid = self.generateImageID()
        thname = self.getThumbName(fname)        
        ret = imgdb.addImage(newid,fname,thname,1)
        if not ret:
            print "Error adding image:",fname
            return 0
        self.images[newid] = [fname,thname]
        return newid

    def addDir(self,dirn,recurse = 0, yieldFcn = None):
        assert(dirn)
        added = 0
        if dirn[-1] != os.sep: dirn += os.sep
        if not os.path.exists(dirn):
            return 0
        
        dfiles = os.listdir(dirn)
        if yieldFcn: yieldFcn()
        for fil in dfiles:
            if recurse and os.path.isdir(dirn+fil):
                if yieldFcn: yieldFcn()                
                added += self.addDir(dirn+fil,recurse)
                continue
            ext = fil.rfind(".")
            if ext == -1:
                continue
            else:
                if fil[ext+1:].lower() in self.supported_ext:
                    if self.addFile(dirn+fil):
                        added += 1
        return added

    def queryId(self, id, numres = 20):
        assert(id)
        nres = imgdb.queryDataFromID(numres,9999999,1,id)
        res = []
        for i in range(int(len(nres)/2)):
            rid = nres[i*2]
            rsc = nres[i*2+1]
            rsc = -100.0*rsc/34.50
            #sanity checks
            if rsc<0:rsc = 0
            if rsc>100:rsc = 100
            res.append([int(rid),rsc])
        if not res: return res
        return res        
    
    def querySig(self,sig,numres = 20,thres = 999999):
        """query for similar images. Source is a db image

        @param id: img id
        @param numres: maximum number of results
        @return: list of similar images. List of pairs [result id,result score]  Score is 0-100
        @rtype: list
        """        
        avgl = sig[:3]
        sigs = map(lambda g: int(g),sig[3:])
    
        nres = imgdb.queryDataFromSig(numres,thres,avgl,sigs)
        res = []
        for i in range(int(len(nres)/2)):
            rid = nres[i*2]
            rsc = nres[i*2+1]
            rsc = -100.0*rsc/34.50
            #sanity checks
            if rsc<0:rsc = 0
            if rsc>100:rsc = 100
            res.append([int(rid),rsc])
        if not res: return res        
        return res

    def calcSig(self,img):
        assert(img)
        return imgdb.calcSig(img)
    
    def getThumbPath(self, id):
        assert(self.images.has_key(id))
        return self.images[id][1]

    def getImgPath(self, id):
        assert(self.images.has_key(id))
        return self.images[id][0]

    def getThumbName(self,fname):
        """return the thumbnail full path for the image
        given by fname (fname is full path) """
        return self.thdir2+ md5.new("file://"+fname).hexdigest()+".png"
    
if __name__ == '__main__':
    print "See test*.py for some tests"
