/***************************************************************************
    imgSeek ::  database C/C++ implementation
                             -------------------
    begin                : Fri Jan 17 2003
    email                : nieder|at|mail.ru
    Time-stamp:            <2003-09-20 01:35:08 rnc>

    Copyright (C) 2003 Ricardo Niederberger Cabral
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* STL Includes */
#include <map>
#include <queue>
#include <list>
#include <fstream>
#include <iostream>
#include <string.h>
#include <math.h>

using namespace std;

/* imgSeek includes */
#include "haar.h"
               
double *new_darray(int size) {
  // python array wrapper. Creates a new double array
  return (double*) malloc(size*sizeof(double));
}

int  *new_iarray(int size) {
  // python array wrapper. Creates a new int array
  return (int*) malloc(size*sizeof(int));
}

inline double *absarray(double* a) {
  // calculates absolute values inplace
  double * b=(double*) malloc(16384*sizeof(double));
  int i;
  for (i=0;i<16384;i++) b[i] = fabs(a[i]);
  return b;
}

inline void truncq(double* a,double* b,double limit,int* data) {
  /* set every element on data to 0 (if < limit) or to a positive/negative int when abs(value) > limit.
     a is the absolute value of array b   
  */ 
  int i,c=0;
  memset(data,0,40*sizeof(int));
  for (i=0;i<16384;i++) {
    if (a[i]>limit) {
      if (b[i]>0) data[c]=i;
      else data[c]=-1*i;
      c++;
      if (c==40)
        {
          return;
        }
    }
  }
}

void transformChar(unsigned char* c1,unsigned char* c2,unsigned char* c3,double* a,double* b,double* c) {
  // do the Haar tensorial 2d transform itself
  // here input RGB data is on unsigned char arrays
  int i,h,j,k;
  double * d1=(double*) malloc(16384*sizeof(double));
  double * d2=(double*) malloc(16384*sizeof(double));
  double * d3=(double*) malloc(16384*sizeof(double));
  double * Ab1=(double*) malloc(128*sizeof(double));
  double * Ab2=(double*) malloc(128*sizeof(double));
  double * Ab3=(double*) malloc(128*sizeof(double));
  /* RGB -> YIQ */
  for (i=0;i<16384;i++) {
    d1[i]=(c1[i]*0.299+c2[i]*0.587+c3[i]*0.114)/256.0;
    d2[i]=(c1[i]*0.596+c2[i]*(-0.274)+c3[i]*(-0.322))/256.0;
    d3[i]=(c1[i]*0.212+c2[i]*(-0.523)+c3[i]*0.311)/256.0;
  }
  // decompose rows
  for (i=0;i<128;i++) {
    h=128;
    for (j=0;j<128;j++) {
      d1[i*128+j] /= 11.314;
      d2[i*128+j] /= 11.314;
      d3[i*128+j] /= 11.314;      
    }
    while(h>1) {
      h/=2;
      for (k=0;k<h;k++) {
        Ab1[k]=(d1[i*128+2*k]+d1[i*128+2*k+1])/1.414;
        Ab2[k]=(d2[i*128+2*k]+d2[i*128+2*k+1])/1.414;
        Ab3[k]=(d3[i*128+2*k]+d3[i*128+2*k+1])/1.414;
        
        Ab1[k+h]=(d1[i*128+2*k]-d1[i*128+2*k+1])/1.414;
        Ab2[k+h]=(d2[i*128+2*k]-d2[i*128+2*k+1])/1.414;
        Ab3[k+h]=(d3[i*128+2*k]-d3[i*128+2*k+1])/1.414;
      }
      memcpy(d1+i*128,Ab1,2*h*sizeof(double));
      memcpy(d2+i*128,Ab2,2*h*sizeof(double));
      memcpy(d3+i*128,Ab3,2*h*sizeof(double));
    }
  }
  // decompose cols
  for (i=0;i<128;i++) {
    h=128;
    for (j=0;j<128;j++) {
      d1[j*128+i] /= 11.314;
      d2[j*128+i] /= 11.314;
      d3[j*128+i] /= 11.314;      
    }
    while(h>1) {
      h/=2;
      for (k=0;k<h;k++) {
        Ab1[k]=(d1[i+2*k*128]+d1[i+(2*k+1)*128])/1.414;
        Ab2[k]=(d2[i+2*k*128]+d2[i+(2*k+1)*128])/1.414;
        Ab3[k]=(d3[i+2*k*128]+d3[i+(2*k+1)*128])/1.414;
        
        Ab1[k+h]=(d1[i+2*k*128]-d1[i+(2*k+1)*128])/1.414;
        Ab2[k+h]=(d2[i+2*k*128]-d2[i+(2*k+1)*128])/1.414;
        Ab3[k+h]=(d3[i+2*k*128]-d3[i+(2*k+1)*128])/1.414;
      }
      for (j=0;j<2*h;j++) {
        d1[j*128+i]=Ab1[j];
        d2[j*128+i]=Ab2[j];
        d3[j*128+i]=Ab3[j];
      }
    }
  }
  memcpy(a,d1,16384*sizeof(double));
  memcpy(b,d2,16384*sizeof(double));
  memcpy(c,d3,16384*sizeof(double));
  free(d1);free(d2);free(d3);free(Ab1);free(Ab2);free(Ab3);

}

void transform(double* a,double* b,double* c) {
  // do the Haar tensorial 2d transform itself
  // here input RGB data is on double char arrays
  int i,h,j,k;
  double * d1=(double*) malloc(16384*sizeof(double));
  double * d2=(double*) malloc(16384*sizeof(double));
  double * d3=(double*) malloc(16384*sizeof(double));
  double * Ab1=(double*) malloc(128*sizeof(double));
  double * Ab2=(double*) malloc(128*sizeof(double));
  double * Ab3=(double*) malloc(128*sizeof(double));
  /* RGB -> YIQ */
  for (i=0;i<16384;i++) {
    d1[i]=(a[i]*0.299+b[i]*0.587+c[i]*0.114)/256.0;
    d2[i]=(a[i]*0.596+b[i]*(-0.274)+c[i]*(-0.322))/256.0;
    d3[i]=(a[i]*0.212+b[i]*(-0.523)+c[i]*0.311)/256.0;
  }
  // decompose rows
  for (i=0;i<128;i++) {
    h=128;
    for (j=0;j<128;j++) {
      d1[i*128+j] /= 11.314;
      d2[i*128+j] /= 11.314;
      d3[i*128+j] /= 11.314;      
    }
    while(h>1) {
      h/=2;
      for (k=0;k<h;k++) {
        Ab1[k]=(d1[i*128+2*k]+d1[i*128+2*k+1])/1.414;
        Ab2[k]=(d2[i*128+2*k]+d2[i*128+2*k+1])/1.414;
        Ab3[k]=(d3[i*128+2*k]+d3[i*128+2*k+1])/1.414;
        
        Ab1[k+h]=(d1[i*128+2*k]-d1[i*128+2*k+1])/1.414;
        Ab2[k+h]=(d2[i*128+2*k]-d2[i*128+2*k+1])/1.414;
        Ab3[k+h]=(d3[i*128+2*k]-d3[i*128+2*k+1])/1.414;
      }
      memcpy(d1+i*128,Ab1,2*h*sizeof(double));
      memcpy(d2+i*128,Ab2,2*h*sizeof(double));
      memcpy(d3+i*128,Ab3,2*h*sizeof(double));
    }
  }
  // decompose cols
  for (i=0;i<128;i++) {
    h=128;
    for (j=0;j<128;j++) {
      d1[j*128+i] /= 11.314;
      d2[j*128+i] /= 11.314;
      d3[j*128+i] /= 11.314;      
    }
    while(h>1) {
      h/=2;
      for (k=0;k<h;k++) {
        Ab1[k]=(d1[i+2*k*128]+d1[i+(2*k+1)*128])/1.414;
        Ab2[k]=(d2[i+2*k*128]+d2[i+(2*k+1)*128])/1.414;
        Ab3[k]=(d3[i+2*k*128]+d3[i+(2*k+1)*128])/1.414;
        
        Ab1[k+h]=(d1[i+2*k*128]-d1[i+(2*k+1)*128])/1.414;
        Ab2[k+h]=(d2[i+2*k*128]-d2[i+(2*k+1)*128])/1.414;
        Ab3[k+h]=(d3[i+2*k*128]-d3[i+(2*k+1)*128])/1.414;
      }
      for (j=0;j<2*h;j++) {
        d1[j*128+i]=Ab1[j];
        d2[j*128+i]=Ab2[j];
        d3[j*128+i]=Ab3[j];
      }
    }
  }
  memcpy(a,d1,16384*sizeof(double));
  memcpy(b,d2,16384*sizeof(double));
  memcpy(c,d3,16384*sizeof(double));
  free(d1);free(d2);free(d3);free(Ab1);free(Ab2);free(Ab3);
}

int calcHaar(double* cdata1,double* cdata2,double* cdata3,int* sig1,int* sig2,int* sig3,double * avgl)
{
  //TODO: Speed up..
  int cnt,i;
  valStruct vals[41];

  double * cadata1=absarray(cdata1);
  double * cadata2=absarray(cdata2);
  double * cadata3=absarray(cdata3);

  avgl[0]=cdata1[0];
  avgl[1]=cdata2[0];
  avgl[2]=cdata3[0];
  // determines the 40th largest absolute value. For each color channel
  valqueue  vq;
  cnt=0;
  for (i=0;i<16384;i++) {
    if (cnt==40) {
      vals[40].d=cadata1[i];
      vq.push(vals[40]);
      vals[40]=vq.top();
      vq.pop();
    }
    else {
      vals[cnt].d=cadata1[i];
      vq.push(vals[cnt]);      
      cnt++;
    }
  }
  truncq(cadata1,cdata1,vq.top().d,sig1);

  while(!vq.empty()) vq.pop();
  cnt=0;
  for (i=0;i<16384;i++) {
    if (cnt==40) {
      vals[40].d=cadata2[i];
      vq.push(vals[40]);
      vals[40]=vq.top();
      vq.pop();
    }
    else {
      vals[cnt].d=cadata2[i];
      vq.push(vals[cnt]);      
      cnt++;
    }
  }
  truncq(cadata2,cdata2,vq.top().d,sig2);

  while(!vq.empty()) vq.pop();
  cnt=0;
  for (i=0;i<16384;i++) {
    if (cnt==40) {
      vals[40].d=cadata3[i];
      vq.push(vals[40]);
      vals[40]=vq.top();
      vq.pop();
    }
    else {
      vals[cnt].d=cadata3[i];
      vq.push(vals[cnt]);      
      cnt++;
    }
  }
  truncq(cadata3,cdata3,vq.top().d,sig3);

  free(cadata1);
  free(cadata2);
  free(cadata3);
  return 1;
}


/* Database */
#include "imgdb.h"
/* Python interface */
#include "imgdb_wrap.cxx"

/* C Includes */
#include <stdio.h>
/* wxWindows includes */
#include "wx/image.h"
#include "wx/file.h"
#include "wx/mstream.h"
#include "wx/wfstream.h"
#include "wx/quantize.h"

void initImgBin() {
  int i,j;
  for (i = 0;i<128;i++) for (j=0;j<128;j++) imgBin[i*128+j] = min(max(i,j),5);
}
void initDbase() {
  /* should be called before adding images */
  printf("Image database initialized.\n");
  initImgBin();
  wxInitAllImageHandlers();
 
}

void free_sigs() {
  for (sigIterator it = sigs.begin(); it != sigs.end(); it++) {
    free((*it).second->sig1);
    free((*it).second->sig2);
    free((*it).second->sig3);
    free((*it).second->avgl);
    delete (*it).second;
  }
}

void closeDbase() {
  /* should be called before exiting app */
  free_sigs();
  printf("Image database closed.\n");
}

int addImage(const long int id, char* filename,char* thname,int doThumb,int ignDim = 0) {
  /* id is a unique image identifier
     filename is the image location
     thname is the thumbnail location for this image
     doThumb should be set to 1 if you want to save the thumbnail on thname
     Images with a dimension smaller than ignDim are ignored
  */
  double* avgl = (double*)malloc(3*sizeof(double));
  int* sig1;
  int* sig2;
  int* sig3;
  int cn = 0;
  double* cdata1 = new_darray(16384);
  double* cdata2 = new_darray(16384);
  double* cdata3 = new_darray(16384);
  int i;

  sig1 = new_iarray(40);
  sig2 = new_iarray(40);
  sig3 = new_iarray(40);

  sigStruct* nsig = new sigStruct();
  nsig->sig1 = sig1;
  nsig->sig2 = sig2;
  nsig->sig3 = sig3;
  nsig->avgl = avgl;
  nsig->id = id;

  if (sigs.count(id)) {     
    printf("ID already in DB: %d\n",id);   
    delete sigs[id];
    sigs.erase(id);
  }
  wxImage image(filename);
  if (! image.Ok() ) return 0;  

  nsig->width = image.GetWidth();
  nsig->height = image.GetHeight();

  if ((ignDim) && ((nsig->width <= ignDim) || (nsig->height <= ignDim))) return 2;

  if (doThumb) {
    wxImage im2(image);

    if (nsig->width > nsig->height) 
      im2.Rescale(128,(int)((128*nsig->height)/nsig->width));
    else
      im2.Rescale((int)((128*nsig->width)/nsig->height),128);      

    //    im2.Rescale(128,128);      //// REMOVE ME when proper redim is back
    im2.SaveFile(thname,wxBITMAP_TYPE_JPEG);
  }
  image.Rescale(128,128);  
  unsigned char* rchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));
  unsigned char* gchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));
  unsigned char* bchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));

  unsigned char* pixel_cache = image.GetData() ;

  for (int idx = 0;idx<16384;idx++) {
    rchan[idx] = *pixel_cache;
    pixel_cache++;
    gchan[idx] = *pixel_cache;
    pixel_cache++;
    bchan[idx] = *pixel_cache;
    pixel_cache++;
  }

  transformChar(rchan,gchan,bchan,cdata1,cdata2,cdata3);

  free(rchan);
  free(bchan);
  free(gchan);

  sigs[id] = nsig;
  calcHaar(cdata1,cdata2,cdata3,sig1,sig2,sig3,avgl);

  for (i = 0;i<40;i++) {         // populate buckets
    if (sig1[i]>0) imgbuckets[0][0][sig1[i]].push_back(id);
    if (sig1[i]<0) imgbuckets[0][1][-sig1[i]].push_back(id);

    if (sig2[i]>0) imgbuckets[1][0][sig2[i]].push_back(id);
    if (sig2[i]<0) imgbuckets[1][1][-sig2[i]].push_back(id);

    if (sig3[i]>0) imgbuckets[2][0][sig3[i]].push_back(id);
    if (sig3[i]<0) imgbuckets[2][1][-sig3[i]].push_back(id);   
  }

  free(cdata1);
  free(cdata2);
  free(cdata3);

  return 1;
}

int loaddb(char* filename) {
  std::ifstream f(filename,ifstream::binary);
  if (!f.is_open()) return 0;
  int sz,coef,c,k;
  long int id;

  // read buckets
  for ( c = 0;c<3;c++)  for (int pn=0;pn<2;pn++)
    for (int i = 0;i<16384;i++) {
      f.read ((char*)&(sz), sizeof(int) );
      for ( k = 0;k<sz;k++) {
        f.read ((char*)&(id), sizeof(long int) );
        imgbuckets[c][pn][i].push_back(id); 
      }
    }
  // read sigs
  f.read ((char*)&(sz), sizeof(int) );

  for ( k = 0;k<sz;k++) {
    f.read ((char*)&(id), sizeof(long int) );
    sigs[id] = new sigStruct();
    sigs[id]->id = id;
    // cout << id <<endl;
    sigs[id]->sig1 = new_iarray(40);
    sigs[id]->sig2 = new_iarray(40);
    sigs[id]->sig3 = new_iarray(40);
    sigs[id]->avgl = (double*)malloc(3*sizeof(double));
    // sig
    for ( c = 0;c<40;c++) {
      f.read ((char*)&(coef), sizeof( int) );   
      sigs[id]->sig1[c] = coef;
      f.read ((char*)&(coef), sizeof( int) );   
      sigs[id]->sig2[c] = coef;
      f.read ((char*)&(coef), sizeof( int) );   
      sigs[id]->sig3[c] = coef;    
    }
    // avgl
    for ( c = 0;c<3;c++) {
      f.read ((char*)&(sigs[id]->avgl[c]), sizeof( double) );      
    }
    // dims
    f.read ((char*)&(sigs[id]->width), sizeof( int) );
    f.read ((char*)&(sigs[id]->height), sizeof( int) );
  }
  f.close();
  return 1;
}

int savedb(char* filename) {
  /*
    Serialization order:
    for each color {0,1,2}:
    for {positive,negative}:
    for each 128x128 coefficient {0-16384}:
    [int] bucket size (size of list of ids)
    for each id:
    [long int] image id
    [int] number of images (signatures)
    for each image:
    [long id] image id
    for each sig coef {0-39}:  (the 40 greatest coefs)
    for each color {0,1,2}:
    [int] coef index (signed)
    for each color {0,1,2}:
    [double] average luminance
    [int] image width
    [int] image height

  */
  std::ofstream f(filename,ofstream::binary);
  if (!f.is_open()) return 0;
  int sz,c;
  long int id;
  // save buckets
  for ( c = 0;c<3;c++)  for (int pn=0;pn<2;pn++)
    for (int i = 0;i<16384;i++) {
      sz = imgbuckets[c][pn][i].size();
      f.write((char*)&(sz), sizeof(int) );
      for (long_listIterator it = imgbuckets[c][pn][i].begin(); it != imgbuckets[c][pn][i].end(); it++) {
        f.write ((char*)&((*it)), sizeof(long int) );
      }
    }
  // save sigs
  sz = sigs.size();
  f.write ((char*)&(sz), sizeof(int) );
  for (sigIterator it = sigs.begin(); it != sigs.end(); it++) {
    id = (*it).first;
    f.write ((char*)&(id), sizeof(long int));
    // sigs 
    for ( c = 0;c<40;c++) {
      f.write ((char*)&((*it).second->sig1[c]), sizeof( int));
      f.write ((char*)&((*it).second->sig2[c]), sizeof( int));
      f.write ((char*)&((*it).second->sig3[c]), sizeof( int));
    }
    // avgl
    for ( c = 0;c<3;c++)
      f.write ((char*)&((*it).second->avgl[c]), sizeof(double));
    // dimensions
    f.write ((char*)&((*it).second->width), sizeof(int));
    f.write ((char*)&((*it).second->height), sizeof(int));
  }
  f.close();
  return 1;
}

std::vector<double> calcSig(char* filename) {
  double* avgl = (double*)malloc(3*sizeof(double));
  int* sig1;
  int* sig2;
  int* sig3;
  int cn = 0;
  double* cdata1 = new_darray(16384);
  double* cdata2 = new_darray(16384);
  double* cdata3 = new_darray(16384);

  sig1 = new_iarray(40);
  sig2 = new_iarray(40);
  sig3 = new_iarray(40);
  
  wxImage image(filename);
  if (! image.Ok() ) {
    vector<double> V;
    return V;
  }

  image.Rescale(128,128);  
  unsigned char* rchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));
  unsigned char* gchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));
  unsigned char* bchan = (unsigned char*) malloc(128*128*sizeof(unsigned char));

  unsigned char* pixel_cache = image.GetData() ;

  for (int idx = 0;idx<16384;idx++) {
    rchan[idx] = *pixel_cache;
    pixel_cache++;
    gchan[idx] = *pixel_cache;
    pixel_cache++;
    bchan[idx] = *pixel_cache;
    pixel_cache++;
  }

  transformChar(rchan,gchan,bchan,cdata1,cdata2,cdata3);

  free(rchan);
  free(bchan);
  free(gchan);

  calcHaar(cdata1,cdata2,cdata3,sig1,sig2,sig3,avgl);

  free(cdata1);
  free(cdata2);
  free(cdata3);

  vector<double> V;
  int b;
  for ( b = 0; b < 3; b++) V.insert(V.end(), avgl[b]);
  for ( b = 0; b < 40; b++) V.insert(V.end(), sig1[b]);
  for ( b = 0; b < 40; b++) V.insert(V.end(), sig2[b]);
  for ( b = 0; b < 40; b++) V.insert(V.end(), sig3[b]);
  
  return V;

}

// set thres to something huge if you don't want smilarity threhold limits
std::vector<double> queryDataFromSig(int numres, float thres,const std::vector<double>& avgl, const std::vector<int>& psigs) {
  int idx,c;
  bool pn;

  int nres = numres+1;

  for (sigIterator sit = sigs.begin(); sit!=sigs.end(); sit++) { // TODO: do I really need to score every single sig on db?
    (*sit).second->score = 0;
    for (c = 0; c<3; c++) {
      (*sit).second->score += weights[0][0][c]*fabs((*sit).second->avgl[c]-avgl[c]);
    }
  }
  for (int b = 0;b<40;b++) {      // for every coef on a sig
    for ( c = 0;c<3;c++) {
      pn = 0; //TODO: c*40+b could be cached on a var. Benchmark this....
      if (psigs[c*40+b]>0) {
        pn = 0;
        idx = psigs[c*40+b];
      } else {
        pn = 1;
        idx = -psigs[c*40+b];
      }
      // update the score of every image which has this coef
      for (long_listIterator uit = imgbuckets[c][pn][idx].begin(); uit != imgbuckets[c][pn][idx].end(); uit++) {
        sigs[(*uit)]->score -= weights[0][imgBin[idx]][c];
      }
    }
  }
  int cnt = 0;
  for (sigIterator it = sigs.begin(); it != sigs.end(); it++) {
    cnt++;
    if ((*it).second->score < thres) {
      pqResults.push(*(*it).second);
      if (cnt>nres) pqResults.pop();      
    }
  }

  vector<double> V;
  sigStruct curResTmp;            /* current result waiting to be returned */

  while (pqResults.size()) {
    curResTmp = pqResults.top();
    pqResults.pop();
    V.insert(V.end(), curResTmp.id);
    V.insert(V.end(), curResTmp.score);
  }
  
  return V;
}

// set thres to something huge if you don't want smilarity threhold limits
std::vector<double> queryDataFromID(int numres, float thres,int scanned, long int id) {
  int idx,c;
  bool pn;
  int nres = numres+1;

  if (!sigs.count(id)) {
    printf("ID not found: %d\n",id);
    vector<double> V;
    return   V;
  }

  int * sig[3] = {sigs[id]->sig1,sigs[id]->sig2,sigs[id]->sig3};
  for (sigIterator sit = sigs.begin(); sit!=sigs.end(); sit++) { // TODO: do I really need to score every single sig on db?
    (*sit).second->score = 0;
    for (c = 0; c<3; c++) {
      (*sit).second->score += weights[0][0][c]*fabs((*sit).second->avgl[c]-sigs[id]->avgl[c]);
    }
  }
  for (int b = 0;b<40;b++) {      // for every coef on a sig
    for ( c = 0;c<3;c++) {
      pn = 0;
      if (sig[c][b]>0) {
        pn = 0;
        idx = sig[c][b];
      } else {
        pn = 1;
        idx = -sig[c][b];
      }
      // update the score of every image which has this coef
      for (long_listIterator uit = imgbuckets[c][pn][idx].begin(); uit != imgbuckets[c][pn][idx].end(); uit++) {
        sigs[(*uit)]->score -= weights[scanned][imgBin[idx]][c];
      }
    }
  }
  int cnt = 0;
  for (sigIterator it = sigs.begin(); it != sigs.end(); it++) {
    cnt++;
    if ((*it).second->score < thres) {
      pqResults.push(*(*it).second);
      if (cnt>nres) pqResults.pop();      
    }
  }

  vector<double> V;
  sigStruct curResTmp;            /* current result waiting to be returned */
  while (pqResults.size()) {
    curResTmp = pqResults.top();
    pqResults.pop();
    V.insert(V.end(), curResTmp.id);
    V.insert(V.end(), curResTmp.score);
  }
  
  return V;
}

void removeID(long int id) { 
  /*remove image with this id from dbase
   */
  if (!sigs.count(id)) {         // don't remove something which isn't even on db.
    cout << "Attempt to remove invalid id: " << id << endl;
    return;
  }
  free(sigs[id]->sig1);
  free(sigs[id]->sig2);
  free(sigs[id]->sig3);
  free(sigs[id]->avgl);
  delete sigs[id];
  sigs.erase(id); 
  // remove id from each bucket it could be in
  for (int c = 0;c<3;c++)  for (int pn=0;pn<2;pn++)
    for (int i = 0;i<16384;i++) {
      imgbuckets[c][pn][i].remove(id);     
    }
}

double calcAvglDiff(long int id1,long int id2) {
  /* return the average luminance difference*/
  if (!sigs.count(id1)) return 0;
  if (!sigs.count(id2)) return 0;
  return fabs(sigs[id1]->avgl[0]-sigs[id2]->avgl[0])+fabs(sigs[id1]->avgl[1]-sigs[id2]->avgl[1])+fabs(sigs[id1]->avgl[2]-sigs[id2]->avgl[2]); 
}

double calcDiff(long int id1,long int id2) {
  /* use it to tell the content-based difference between two images
   */
  double diff = calcAvglDiff(id1,id2);
  int * sig1[3] = {sigs[id1]->sig1,sigs[id1]->sig2,sigs[id1]->sig3};
  int * sig2[3] = {sigs[id2]->sig1,sigs[id2]->sig2,sigs[id2]->sig3};
  for (int b = 0;b<40;b++) for (int c=0;c<3;c++) for (int b2 = 0;b2<40;b2++) {
    if (sig2[c][b2] == sig1[c][b]) 
      diff -= weights[0][imgBin[abs(sig1[c][b])]][c];
  }  
  return diff;
}

int resetdb(void) {
  /* call it to reset all buckets and signatures 
   */
  for (int c = 0;c<3;c++)  for (int pn=0;pn<2;pn++) for (int i=0;i<16384;i++) imgbuckets[c][pn][i].clear();
  // delete sigs
  free_sigs();
  sigs.clear();
  return 1;
}

